@@ -24,8 +24,9 @@
 #define PIM_IGMP_JOIN_H
 
 #include <netinet/in.h>
+#include "if.h"
 
-int pim_igmp_join_source(int fd, int ifindex,
+int pim_igmp_join_source(int fd, ifindex_t ifindex,
 			 struct in_addr group_addr,
 			 struct in_addr source_addr);
 
